from .formatting import (
    DARK_BLUE,
    BLUE,
    LIGHT_BLUE,
    LIGHTER_BLUE,
    DARK_GREEN,
    LIGHT_GREEN,
    DARK_GRAY,
    GRAY,
    LIGHT_GRAY,
    WHITE,
    format_fig,
    display_fig,
)
