from policyengine_core.errors import EmptyArgumentError

from .abstract_rate_tax_scale import AbstractRateTaxScale
from .abstract_tax_scale import AbstractTaxScale
from .amount_tax_scale_like import AmountTaxScaleLike
from .helpers import combine_tax_scales
from .linear_average_rate_tax_scale import LinearAverageRateTaxScale
from .marginal_amount_tax_scale import MarginalAmountTaxScale
from .marginal_rate_tax_scale import MarginalRateTaxScale
from .rate_tax_scale_like import RateTaxScaleLike
from .single_amount_tax_scale import SingleAmountTaxScale
from .tax_scale_like import TaxScaleLike
