class Unit:
    """The real-world meaning of a particular value."""

    GBP = "currency-GBP"
    """British pounds sterling."""

    USD = "currency-USD"
    """United States dollars."""

    EUR = "currency-EUR"
    """Euros."""

    PERCENT = "/1"
    """A percentage."""
